var path   = require('path')
  , _      = require('underscore')
  , fs     = require('fs')
  , os     = require('os')
  , util   = require('util')
  , xml2js = require('xml2js')
  , exec   = require('child_process').exec
  , spawn   = require('child_process').spawn;

var app = require('../app');
app.set('rootPath', './');
var isMultiuser = module.parent.exports.set("isMultiuser");
var dotPath = "dot";

// onload
var defs = "";

var environment = {};
var rossetup = "";

exec("node createdb.js");

var distribution = os.type();
function updatedefs(onclose, req){
    console.log("updating defs");
    var child;
    var env = process.env.FLOWDESIGNER_PATH || "";

    // distribution decision
    if(distribution.indexOf("Linux") !== -1)   distribution = "Linux";
    if(distribution.indexOf("Windows") !== -1) distribution = "Windows";
    if(distribution.indexOf("Darwin") !== -1)  distribution = "Darwin";
    console.log("OS: " + distribution);

    if(distribution === "Linux" || distribution === "Darwin"){
        paths = env.split(":");
        paths.push("`flowdesigner-config --toolboxdir`");

        if(!_.isUndefined(req)){
            var userpath = process.cwd() + "/" + getUserdirPath(req, isMultiuser) + "_package";
            paths.push(userpath);
        }

        child = exec("find " + paths.join(" ") + " -name '*.def' ");
        child.stdout.on("data", function(data){ defs += "\n" + data + "\n"; });
        child.on("close", function(code, signal){
            defs = _.uniq(defs.split("\n"));
            defs = _.uniq(defs, function(def){return def.split(path.sep).slice(-1)[0];});
            defs = _.sortBy(defs, function(def){return def.split(path.sep).slice(-1)[0];});
            defs = _.filter(defs, function(def){return def;});
            defs = _.chain(defs).map(function(d){return d.split(",");}).flatten().value();
            defs = _.uniq(defs);
            console.log("in Linux or Darwin", util.inspect(defs));
            if(_.isFunction(onclose)) onclose();});
    }
    else if(distribution === "Windows"){
        distribution = "Windows";
        console.log("OS: Windows");
        paths = env.split(";");
        cmd = 'dir /s /b ';
        _.each(paths, function(p){
            cmd += '"' + p + '"';
        });
        cmd += '| findstr -i \\.def$';

        console.log(cmd);
        child = exec(cmd);
        child.stdout.on("data", function(data){ defs += "\r\n" + data + "\r\n"; });
        child.on("close", function(code, signal){
            defs = _.uniq(defs.split("\r\n"));
            defs = _.uniq(defs, function(def){return def.split(path.sep).slice(-1)[0];});
            defs = _.sortBy(defs, function(def){return def.split(path.sep).slice(-1)[0];});
            defs = _.filter(defs, function(def){return def;});
            defs = _.chain(defs).map(function(d){return d.split(",");}).flatten().value();
            defs = _.uniq(defs);
            console.log("found defs:", defs);
            if(_.isFunction(onclose)) onclose();
        });

        // find dot (graphviz binary). seems ugly. sync version of exec is needed.
        exec("dot", function(err, stdout, stderr){
            if(err){
                var child1 = exec('dir /s /b "C:\\Program Files" | findstr -i dot.exe | findstr -i graphviz', function(err, stdout, stderr){
                    if(err){
                        var child2 = exec('dir /s /b "C:\\Program Files (x86)" | findstr -i dot.exe | findstr -i graphviz', function(err, stdout, stderr){
                            if(err){
                                dotPath = false;
                            } else{
                                dotPath = '"' + stdout.slice(0, -2) + '"';
                                console.log("path to graphviz executable:", dotPath);
                            }
                        });
                    }
                    else{
                        dotPath = '"' + stdout.slice(0, -2) + '"';
                        console.log("path to graphviz executable:", dotPath);
                    }
                });
            }else{
                dotPath = "dot";
                console.log("path to graphviz executable:", dotPath);
            }
        });
    }
}

updatedefs();

exports.index = function (req, res) {
    if(isMultiuser &&
       (_.isUndefined(req.session) || _.isUndefined(req.session.user))) {
        console.log("exports.index, redirect, req.session=", req.session);
        res.redirect("/login");
    }
    updatedefs(undefined, req);

    console.log("root:", req.query.root);
    if(req.query.root === "hark-designer")
        app.set(rootPath, "hark-designer/");

    selenium = "right";
    if(req.query.selenium === "selenium")
        selenium = "left";
    // res.render('index', { title: 'Express' });
    // res.sendfile('public/index.html');
    var css = ['stylesheetslib/smoothness/jquery-ui-1.10.2.custom.min.css',
               'stylesheetslib/jquery.contextMenu.css',
               'stylesheetslib/introjs.min.css',
               'stylesheets/main.min.css'];
    var js =  [
               'jslib/json2.js',
               'jslib/jquery.min.1.9.1.js',
               'jslib/jquery-ui-1.10.2.custom.min.js',
               'jslib/jquery.ui.position.js',
               'jslib/jquery.contextMenu.js',
               'jslib/underscore-min.js',
               'jslib/backbone-min.js',
               'jslib/intro.min.js',
               'jslib/fileBrowse.js',
               'jslib/jQuery-File-Upload/js/vendor/jquery.ui.widget.js',
               'jslib/jQuery-File-Upload/js/jquery.iframe-transport.js',
               'jslib/jQuery-File-Upload/js/jquery.fileupload.js',
               'jslib/jQuery-File-Upload/js/jquery.fileupload-jquery-ui.js',
               'js/arrayObjectMixin.js',
               'js/util.min.js',
               'js/main.min.js',
               ];

    css = _.map(css, function(tmp){return app.get('rootPath') + tmp;});
    js  = _.map(js , function(tmp){return app.get('rootPath') + tmp;});
    res.render('index', {css: css, javascripts: js,
        rootPath: app.get('rootPath'), selenium: selenium,
        session:req.session, isMultiuser: isMultiuser});
};

exports.packagelist = function(req, res){
    res.send(_.chain(defs)
              .map(function(d){return d.split(path.sep).slice(-1)[0];})
              .filter(function(d){return d.indexOf("def") !== -1;}).value());
};

var isDir = function( filepath ) {
    if( /\/$/.test( filepath ) ) {
        filepath = filepath.substr( 0, filepath.lastIndexOf('/')-1 );
    }
    return (fs.existsSync( filepath ) && fs.statSync( filepath ).isDirectory());
};
exports.preferenceSet = function(req, res){
    if(req.params.file === "environment"){
        console.log(req.body);
        if(req.body.rosVersionInternal){
            rossetup = ". /opt/ros/" + req.body.rosVersionInternal + "/setup.sh;";
        }
        delete req.body.rosVersionInternal;
        environment = req.body;
        res.send(200);
        return;
    }
    res.send(404);
}

exports.preference = function(req, res){
    // if(!_.isUndefined(req.session) && !_.isUndefined(req.session.user)){
        // userpath = req.session.user;
    // }

    // path to user package
    var userpath = process.cwd() + "/" + getUserdirPath(req, isMultiuser) + "_package";
    console.log("=========");
    console.log(userpath);
    if(process.env.FLOWDESIGNER_PATH && process.env.FLOWDESIGNER_PATH.indexOf(userpath) === -1){
        // process.env.FLOWDESIGNER_PATH += ":" + userpath;
    } else{
        // process.env.FLOWDESIGNER_PATH = userpath;
    }
    // console.log(process.env.FLOWDESIGNER_PATH.indexOf(userpath));
    console.log("FLOWDESIGNER_PATH:", process.env.FLOWDESIGNER_PATH);
    console.log("=========");


    if(req.params.file === "packages"){
        defs = "";
        updatedefs(
            function(){
                res.render(req.params.file, {
                    files: _.chain(defs)
                            .map(function(d){return d.split(path.sep).slice(-1)[0];})
                            .filter(function(d){return d.indexOf("def") !== -1;}).value(),
                    colors: ["Classic", "Sea", "Fire", "Dark"]
                });
            }, req
        );
    }
    else if(req.params.file === "version"){
        console.log(rossetup + "batchflow -v");
        var envkeys = _.keys(environment);
        var runenv = process.env;
        for(var i=0; i<envkeys.length; i++){
            if(runenv[envkeys[i]]){
                runenv[envkeys[i]] += ":" + environment[envkeys[i]];
                console.log(runenv[envkeys[i]]);
            } else{
                runenv[envkeys[i]] = environment[envkeys[i]];
            }
        }
        console.log("env", runenv);
        exec(rossetup + "batchflow -v", {env: runenv}, function(err, stdout, stderr){
            if(err !== null)
                res.send(200, "batchflow is not installed");
            else{
                res.send(200, stdout.replace("\n", "<br>"));
            }
        });
    }
    else if(req.params.file === "environment"){
        var ros = "undefined";
        if(fs.existsSync("/opt/ros")){
            console.log(fs.readdirSync("/opt/ros/"));
            ros = _.filter(fs.readdirSync("/opt/ros/"), function(f){return isDir("/opt/ros/" + f);});
            console.log(ros);
        }
        console.log("ros", ros);
        res.render(req.params.file, {ros: ros, isMultiuser: isMultiuser});
    }
    else {
        res.render(req.params.file);
    }
};



exports.templates = function(req, res){
    res.render(req.params.name);
};

exports.nodedefs = function(req, res){
    var def = _.find(defs, function(d){
        return d.split(path.sep).slice(-1).indexOf(req.params.name) !== -1;});

    if(def){
        // send processed data
        fs.readFile(def, function(err, data){
            if(err){
                res.send(500);
                return;
            }
            var parser = new xml2js.Parser();
            parser.addListener('end', function(result){
                var defobj = _.map(result.Definitions.NodeClass, def2node);
                // console.log(defs)

                // read .n files in .def files (#1214)
                var dirname = def.replace(def.split(path.sep).slice(-1)[0], "");
                var nfiles = _.filter(fs.readdirSync(dirname),
                                      function(f){return f.slice(-2) === ".n";});
                // console.log("readDirSync:", dirname, nfiles);

                if(nfiles.length > 0){
                    var send = _.after(nfiles.length,
                        function(){res.send(200, JSON.stringify(defobj));});
                    _.each(nfiles, function(nfile){
                        var string = fs.readFileSync(path.join(dirname, nfile), 'utf8');
                        string = string.slice(string.indexOf("\n"));
                        var tmpparser = new xml2js.Parser();
                        tmpparser.addListener('end', function(result){
                            defobj.push(nfile2node(result.Document, nfile.split(".")[0]));
                            defobj = _.sortBy(defobj, "name");
                            send();
                        });
                        tmpparser.parseString(string);
                    });
                }
                else{
                    defobj = _.sortBy(defobj, "name");
                    res.send(200, JSON.stringify(defobj));
                }
            });
            parser.parseString(data);
        });
    }
    else{
        console.log("def not found");
        res.send(404);
    }
};
function nfile2node(nfile, filename){
    var node = {
        category: nfile.$.category,
        description: "",
        inputs: [],
        outputs:[],
        name: filename,
        parameters: []
    };

    node.inputs = _.map(nfile.Network[0].NetInput, function(inp){
        return {
            description: "",
            name: inp.$.name,
            object_type: "any"
        };
    });

    node.outputs = _.map(nfile.Network[0].NetOutput, function(out){
        return {
            description: "",
            name: out.$.name,
            object_type: "any"
        };
    });

    node.parameters = _.map(nfile.Parameter, function(tmp){
        return {
            description: "",
            condition: "",
            list: [],
            name: tmp.$.name,
            type: (tmp.$.type === "")? "string" : tmp.$.type,
            value: tmp.$.value
        };
    });
    return node;
}

function def2node(def){
    var node = {
        category: def.$.category,
        description: def.Description[0],
        inputs: [],
        outputs:[],
        name: def.$.name,
        parameters: []
    };
    node.inputs = _.map(def.Input, function(tmp){
        return { description:tmp._,
                 name:tmp.$.name,
                  object_type: tmp.$.type
                };
    }) || [];

    node.outputs = _.map(def.Output, function(tmp){
        return { description:tmp._,
                 name:tmp.$.name,
                  object_type: tmp.$.type
                };
    }) || [];

    node.parameters = _.map(def.Parameter, function(tmp){
        var typeval = (tmp.$.type === "")? "string" : tmp.$.type;
        var listval = (tmp.$.list === "")? [] : tmp.$.list.split(":");
        if(typeval === "bool") listval = []; // ignore list if type == bool. #1330

        return {
            description: tmp._,
            condition: tmp.$.valid,
            list: listval,
            name: tmp.$.name,
            type: typeval,
            value: tmp.$.value
        };
    });

    return node;
}


var child;
var child_result = "";
var stderr = "";
var stdout = "";
var whole = "";
exports.execution = function(req, res){
    if(fs.existsSync("batchflow")){
        res.send(504, "Flow Designer is not installed");
        return;
    }

    if(child !== undefined){
        console.log("child is not undefined");
        res.send(504, "Another process already exists");
        return;
    }

    stderr = "";
    stdout = "";
    whole = "";
    console.log(getUserdirPath(req, isMultiuser));


    // user added environment variables
    var runenv = process.env;
    var envkeys = _.keys(environment);
    for(var i=0; i<envkeys.length; i++){
        if(runenv[envkeys[i]]){
            runenv[envkeys[i]] += ":" + environment[envkeys[i]];
            console.log(runenv[envkeys[i]]);
        } else{
            runenv[envkeys[i]] = environment[envkeys[i]];
        }
    }

    // add user package path
    // var userpath = process.cwd() + "/" + getUserdirPath(req, isMultiuser) + "_package";
    // if(_.indexOf(envkeys, "FLOWDESIGNER_PATH") === -1){
    //     runenv["FLOWDESIGNER_PATH"] = userpath;
    // } else {
    //     runenv["FLOWDESIGNER_PATH"] += ":" + userpath;
    // }
    // console.log('runenv["FLOWDESIGNER_PATH"]=', runenv["FLOWDESIGNER_PATH"]);

    // environment funcions
    var rosenv = {};
    exec("env", function(error, stdout, stderr){
        _.each(stdout.split("\n"), function(pair){
            var envname = pair.slice(0, pair.indexOf("="));
            var envval  = pair.slice(pair.indexOf("=") + 1);
            rosenv[envname] = envval;
        });

        console.log(rossetup, "env");
        exec(rossetup + "env", function(error, stdout, stderr){
            console.log("ROS env set", stdout);
            _.each(stdout.split("\n"), function(pair){
                var envname = pair.slice(0, pair.indexOf("="));
                var envval  = pair.slice(pair.indexOf("=") + 1);

                // if name exists and value does not change, remove it.
                if(_.indexOf(_.keys(rosenv), envname) !== -1){
                    if(rosenv[envname] !== envval){
                        console.log(envname, "  changed");
                        rosenv[envname] = envval;
                    }
                    else{
                        delete rosenv[envname];
                    }
                }
            });
            _.each(_.keys(rosenv), function(e){
                console.log("ROSENV: ", e, rosenv[e]);
                runenv[e] = rosenv[e];
            });
            // console.log("env", runenv);

            var cmd = 'batchflow ' + req.params.file;
            child = spawn (cmd.split(" ")[0], cmd.split(" ").slice(1),
                          {env       : runenv,
                           cwd       : getUserdirPath(req, isMultiuser),
                           killSignal: "SIGKILL"});
            child.stderr.on("data", function(data){
                stderr += data + "\n"; whole += "<stderr>" + data + "</stderr>\n";
            });
            child.stdout.on("data", function(data){
                stdout += data + "\n"; whole += "<stdout>" + data + "</stdout>\n";
            });
            child.on("close", function(code, signal){
                if(code === 0)
                    child_result = "Normally finished.";
                else
                    child_result = "Stopped";
                child = undefined;
            });

            child.unref();
            res.send(200, "Process started");
        });
    });
};

exports.execstatus = function(req, res){
    if(child !== undefined)
        res.send(200, "Process is running.");
    else
        res.send(200, child_result + "\n" + whole);
};

exports.abort = function(req, res){
    if(child){
        console.log(child.pid);
        child.kill('SIGKILL');
        res.send(200, "Killed " + whole);
        child = undefined;
    }
    else{
        res.send(404, "No process is running");
    }
};

exports.os = function(req, res){
   res.send(200, distribution);
};

exports.keepalive = function(req, res){
    res.send(200);
};


exports.reshape = function(req, res){
    if(!dotPath){
        console.log(dotPath);
        res.send(404, "dot not found. set PATH variable");
        return;
    }

    var network = JSON.parse(JSON.stringify(req.body));
    var filename = "dot/test.dot";

    var ret;
    if(distribution === "Linux" || distribution === "Darwin")   ret = "\n";
    if(distribution === "Windows") ret = "\r\n";

    //convert from network to dot language
    fs.writeFileSync(filename, "");
    fs.appendFileSync(filename, "digraph network {" + ret);
    fs.appendFileSync(filename, "  graph [rankdir = LR]" + ret);
    _.each(network.sheets, function(sheet){
        // filename = "dot/" + sheet.name + ".dot";
        fs.appendFileSync(filename, "  subgraph cluster" + sheet.name + "{" + ret);
        fs.appendFileSync(filename, "    label=\"" + sheet.name + "\";" + ret);
        _.chain(network.nodes).filter(function(n){return n.sheet === sheet.name;})
         .each(function(n){
            fs.appendFileSync(filename, "    " + n.id + "[" + ret);
            fs.appendFileSync(filename, "      shape = Mrecord," + ret);
            fs.appendFileSync(filename, "      label = \"" + n.id + "|");
            fs.appendFileSync(filename, "{{");
            _.each(n.inputs, function(io){
                fs.appendFileSync(filename, util.format("<%s> %s", io.name , io.name));
                if(io !== n.inputs.slice(-1)[0])
                    fs.appendFileSync(filename, "|");
            });
            fs.appendFileSync(filename, "}|{");
            _.each(n.outputs, function(io){
                fs.appendFileSync(filename, util.format("<%s> %s", io.name, io.name));
                if(io !== n.outputs.slice(-1)[0])
                    fs.appendFileSync(filename, "|");
            });
            fs.appendFileSync(filename, "}}");
            fs.appendFileSync(filename, "\"" + ret + "    ]" + ret);
         });

        _.chain(network.links).filter(function(l){return l.sheet === sheet.name;})
         .each(function(l){
            fs.appendFileSync(filename,
                util.format("    %s:%s -> %s:%s;%s",
                            l.fromNodeId, l.fromTermName,
                            l.toNodeId, l.toTermName, ret));
         });
        fs.appendFileSync(filename, "  }\n");
    });
    fs.appendFileSync(filename, "}\n");

    // calculate positions and send it.
    var re = new RegExp(/([a-zA-Z0-9_]+)\W+\[.*pos="(\d+),(\d+)".*\]/);
    exec(dotPath + " dot/test.dot", function(error, stdout, stderr){
        if(error){
            console.log(stderr, error);
            res.send(504, stderr);
        }
        // copy the positions calculated by graphviz
        _.each(stdout.replace(/\n/g, "").replace(/\r/g, "").split(";"), function(line){
            if(line.match("node_") && line.indexOf("->") === -1){
                // console.log(line);
                var result = re.exec(line);
                var nodeid = result[1];
                var left = parseInt(result[2], 10);
                var top  = parseInt(result[3], 10);
                var nodeobj = _.find(network.nodes, function(n){return n.id === nodeid;});
                nodeobj.pos = [left, top];
                // console.log(nodeid, nodeobj.pos);
            }
        });

        // flipud
        var maxval = _.chain(network.nodes).map(function(n){return n.pos[1];}).max().value();
        _.each(network.nodes, function(n){
            n.pos[1] = maxval - n.pos[1];
        });

        // shift
        _.each(_.keys(req.body.sheets), function(sheet){
            // console.log(sheet);
            var minx = _.chain(network.nodes)
                        .filter(function(n){return n.sheet === sheet;})
                        .map(function(n){return n.pos[0];})
                        .min().value();
            var miny = _.chain(network.nodes)
                        .filter(function(n){return n.sheet === sheet;})
                        .map(function(n){return n.pos[1];})
                        .min().value();
            // console.log(minx, miny);
            _.chain(network.nodes).filter(function(n){return n.sheet === sheet;})
             .each(function(n){
                n.pos[0] = truncate((n.pos[0] - minx) * 1.0);
                n.pos[1] = truncate((n.pos[1] - miny) * 1.0);
             });
        });

        // for debug
        _.each(network.nodes, function(n){
            console.log(n.id, n.pos[0], n.pos[1]);
        });
        // console.log(network);

        // restore undefined properties if not exist
        network.links = network.links || [];
        network.terminalRoles = network.terminalRoles || [];
        _.each(network.nodes, function(n){
            _.each(n.parameters, function(p){
                p.list = p.list || [];
            });
        });
        fs.writeFileSync("dot/reshaped.dot", stdout);
        res.send(200, network);
    });
};

function truncate(value){
    return parseInt(value/10, 10) * 10;
}

function getUserdirPath(req, isMultiuser){
    if(isMultiuser){
        return "userdata/" + req.session.user;
    } else{
        return "userdata/networks";
    }
}
